import os
from dataclasses import dataclass
from enum import Enum
from typing import Optional

VYPER_COLOR_OUTPUT = os.environ.get("VYPER_COLOR_OUTPUT", "0") == "1"
VYPER_ERROR_CONTEXT_LINES = int(os.environ.get("VYPER_ERROR_CONTEXT_LINES", "1"))
VYPER_ERROR_LINE_NUMBERS = os.environ.get("VYPER_ERROR_LINE_NUMBERS", "1") == "1"

VYPER_TRACEBACK_LIMIT: Optional[int]

_tb_limit_str = os.environ.get("VYPER_TRACEBACK_LIMIT")
if _tb_limit_str is not None:
    VYPER_TRACEBACK_LIMIT = int(_tb_limit_str)
else:
    VYPER_TRACEBACK_LIMIT = None


class OptimizationLevel(Enum):
    NONE = 1
    GAS = 2
    CODESIZE = 3

    @classmethod
    def from_string(cls, val):
        match val:
            case "none":
                return cls.NONE
            case "gas":
                return cls.GAS
            case "codesize":
                return cls.CODESIZE
        raise ValueError(f"unrecognized optimization level: {val}")

    @classmethod
    def default(cls):
        return cls.GAS


@dataclass
class Settings:
    compiler_version: Optional[str] = None
    optimize: Optional[OptimizationLevel] = None
    evm_version: Optional[str] = None
    experimental_codegen: Optional[bool] = None


_DEBUG = False


def _is_debug_mode():
    global _DEBUG
    return _DEBUG


def _set_debug_mode(dbg: bool = False) -> None:
    global _DEBUG
    _DEBUG = dbg



from vyper.utils import ir_pass
from vyper.venom.analysis import calculate_cfg
from vyper.venom.basicblock import IRInstruction, IRLabel
from vyper.venom.function import IRFunction


def _optimize_unused_variables(ctx: IRFunction) -> set[IRInstruction]:
    """
    Remove unused variables.
    """
    removeList = set()
    for bb in ctx.basic_blocks:
        for i, inst in enumerate(bb.instructions[:-1]):
            if inst.volatile:
                continue
            if inst.output and inst.output not in bb.instructions[i + 1].liveness:
                removeList.add(inst)

        bb.instructions = [inst for inst in bb.instructions if inst not in removeList]

    return removeList


def _optimize_empty_basicblocks(ctx: IRFunction) -> int:
    """
    Remove empty basic blocks.
    """
    count = 0
    i = 0
    while i < len(ctx.basic_blocks):
        bb = ctx.basic_blocks[i]
        i += 1
        if len(bb.instructions) > 0:
            continue

        replaced_label = bb.label
        replacement_label = ctx.basic_blocks[i].label if i < len(ctx.basic_blocks) else None
        if replacement_label is None:
            continue

        # Try to preserve symbol labels
        if replaced_label.is_symbol:
            replaced_label, replacement_label = replacement_label, replaced_label
            ctx.basic_blocks[i].label = replacement_label

        for bb2 in ctx.basic_blocks:
            for inst in bb2.instructions:
                for op in inst.operands:
                    if isinstance(op, IRLabel) and op.value == replaced_label.value:
                        op.value = replacement_label.value

        ctx.basic_blocks.remove(bb)
        i -= 1
        count += 1

    return count


@ir_pass
def ir_pass_optimize_empty_blocks(ctx: IRFunction) -> int:
    changes = _optimize_empty_basicblocks(ctx)
    calculate_cfg(ctx)
    return changes


@ir_pass
def ir_pass_remove_unreachable_blocks(ctx: IRFunction) -> int:
    return ctx.remove_unreachable_blocks()


@ir_pass
def ir_pass_optimize_unused_variables(ctx: IRFunction) -> int:
    return len(_optimize_unused_variables(ctx))



class IRPass:
    """
    Decorator for IR passes. This decorator will run the pass repeatedly
    until no more changes are made.
    """

    @classmethod
    def run_pass(cls, *args, **kwargs):
        t = cls()
        count = 0

        while True:
            changes_count = t._run_pass(*args, **kwargs) or 0
            count += changes_count
            if changes_count == 0:
                break

        return count

    def _run_pass(self, *args, **kwargs):
        raise NotImplementedError(f"Not implemented! {self.__class__}.run_pass()")
